# Andrey version 1.0.0
# AndreyFlattener.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Andrey, a pedagogical implementation of Markov chains.
# Andrey is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Andrey is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Andrey.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Instantiation of models
# 4) Flattening of models

# 1) Imported modules
# -------------------

import string
import sys
import re

from S2MLMetaModel import *
from S2MLFlattener import *
from AndreyModels import *
from AndreyPrinter import *

# 2) Creator
# ----------

class AndreyFlattener(S2MLFlattener):
  def __init__(self, engine):
    S2MLFlattener.__init__(self, engine)

# 3) Instantiation of elements
# ----------------------------

  def NewModel(self):
    return AndreyModel()

  def InstantiateElement(self, action):
    sourceElement = action.GetDatum(0)
    if AndreyModel.IsMarkovChainBlock(sourceElement):
      status = self._InstantiateMarkovChainBlock(action)
    elif AndreyModel.IsMarkovSystemBlock(sourceElement):
      status = self._InstantiateMarkovSystemBlock(action)
    elif AndreyModel.IsMarkovChainState(sourceElement):
      status = self._InstantiateMarkovChainState(action)
    elif AndreyModel.IsMarkovChainEvent(sourceElement):
      status = self._InstantiateMarkovChainEvent(action)
    elif AndreyModel.IsMarkovChainParameter(sourceElement):
      status = self._InstantiateMarkovChainParameter(action)
    elif AndreyModel.IsMarkovChainReward(sourceElement):
      status = self._InstantiateMarkovChainReward(action)
    elif AndreyModel.IsMarkovChainTransition(sourceElement):
      status = self._InstantiateMarkovChainTransition(action)
    else:
      status = S2MLFlattener.InstantiateElement(self, action)
    return status

  def _InstantiateMarkovChainBlock(self, action):
    sourceBlock = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetBlock = self._targetModel.NewBlock(sourceBlock.GetIdentifier(), targetContainer)
    targetBlock.SetSubType(AndreyType.MARKOV_CHAIN)
    self._PushInstantiateActionsForElements(sourceBlock, targetBlock)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateMarkovSystemBlock(self, action):
    sourceBlock = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetBlock = self._targetModel.NewBlock(sourceBlock.GetIdentifier(), targetContainer)
    targetBlock.SetSubType(AndreyType.MARKOV_SYSTEM)
    self._PushInstantiateActionsForElements(sourceBlock, targetBlock)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateMarkovChainState(self, action):
    sourceState = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceState.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        targetState = self._targetModel.NewMarkovChainState(targetPath, targetContainer)
        targetState.Rename(targetPath.ToString())
        context = S2MLPath.NewEmptyPath()
      else:
        absoluteSourcePath = sourceState.GetAbsolutePath()
        self.RaiseError("State %s declared outside of its parent block\n" % absoluteSourcePath.ToString())
        return S2MLFlattenerAction.FAILED
    elif AndreyModel.IsMarkovChainState(targetElement):
      targetState = targetElement # self._targetModel.NewMarkovChainState(targetPath, targetContainer)
      context = targetContainer.GetContextualizedAbsolutePath()
    else:
      absoluteSourcePath = sourceState.GetAbsolutePath()
      self.RaiseError("Type mismatch for state %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    for name in sourceState.GetAttributeNames():
      sourceExpression = sourceState.GetAttribute(name)
      targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, context)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetState.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateMarkovChainEvent(self, action):
    sourceEvent = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceEvent.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        targetEvent = self._targetModel.NewMarkovChainEvent(targetPath, targetContainer)
        targetEvent.Rename(targetPath.ToString())
        context = S2MLPath.NewEmptyPath()
      else:
        absoluteSourcePath = sourceEvent.GetAbsolutePath()
        self.RaiseError("Event %s declared outside of its parent block\n" % absoluteSourcePath.ToString())
        return S2MLFlattenerAction.FAILED
    elif AndreyModel.IsMarkovChainEvent(targetElement):
      targetEvent = targetElement # self._targetModel.NewMarkovChainEvent(targetPath, targetContainer)
      context = targetContainer.GetContextualizedAbsolutePath()
    else:
      absoluteSourcePath = sourceEvent.GetAbsolutePath()
      self.RaiseError("Type mismatch for event %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    for name in sourceEvent.GetAttributeNames():
      sourceExpression = sourceEvent.GetAttribute(name)
      targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, context)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetEvent.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateMarkovChainParameter(self, action):
    sourceParameter = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceParameter.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        targetParameter = self._targetModel.NewMarkovChainParameter(targetPath, targetContainer)
        targetParameter.SetSubType(AndreyType.MARKOV_CHAIN_PARAMETER)
        targetParameter.Rename(targetPath.ToString())
        context = S2MLPath.NewEmptyPath()
      else:
        absoluteSourcePath = sourceParameter.GetAbsolutePath()
        self.RaiseError("Parameter %s declared outside of its parent block\n" % absoluteSourcePath.ToString())
        return S2MLFlattenerAction.FAILED
    elif AndreyModel.IsMarkovChainParameter(targetElement):
      targetParameter = targetElement # AndreyMarkovChainParameter.NewParameter(targetPath, targetContainer)
      context = targetContainer.GetContextualizedAbsolutePath()
    else:
      absoluteSourcePath = sourceParameter.GetAbsolutePath()
      self.RaiseError("Type mismatch for parameter %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    sourceExpression = sourceParameter.GetDefinition()
    targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, context)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetParameter.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateMarkovChainReward(self, action):
    sourceReward = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceReward.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        targetReward = self._targetModel.NewMarkovChainParameter(targetPath, targetContainer)
        targetReward.SetSubType(AndreyType.MARKOV_CHAIN_REWARD)
        targetReward.Rename(targetPath.ToString())
        context = S2MLPath.NewEmptyPath()
      else:
        absoluteSourcePath = sourceReward.GetAbsolutePath()
        self.RaiseError("Reward %s declared outside of its parent block\n" % absoluteSourcePath.ToString())
        return S2MLFlattenerAction.FAILED
    elif AndreyModel.IsMarkovChainReward(targetElement):
      targetReward = targetElement # self._targetModel.NewMarkovChainParameter(targetPath, targetContainer)
      context = targetContainer.GetContextualizedAbsolutePath()
    else:
      absoluteSourcePath = sourceReward.GetAbsolutePath()
      self.RaiseError("Type mismatch for reward %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    sourceExpression = sourceReward.GetDefinition()
    targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, context)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetReward.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateMarkovChainTransition(self, action):
    sourceTransition = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceEventPath = sourceTransition.GetEventPath()
    targetEventPath = S2MLPath.Copy(sourceEventPath)
    targetEvent = targetContainer.AccessNamedElement(targetEventPath)
    if targetEvent==None:
      self.RaiseError("In transition, undeclared event %s\n" % targetEventPath.ToString())
      return S2MLFlattenerAction.FAILED
    if not AndreyModel.IsMarkovChainEvent(targetEvent):
      absolutePath = targetEvent.GetAbsolutePath()
      self.RaiseError("In transition, type mismatch for event %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    sourceSourceStatePath = sourceTransition.GetSourceStatePath()
    targetSourceStatePath = S2MLPath.Copy(sourceSourceStatePath)
    targetSourceState = targetContainer.AccessNamedElement(targetSourceStatePath)
    if targetSourceState==None:
      self.RaiseError("In transition, undeclared state %s\n" % targetSourceStatePath.ToString())
      return S2MLFlattenerAction.FAILED
    if not AndreyModel.IsMarkovChainState(targetSourceState):
      absolutePath = targetSourceState.GetAbsolutePath()
      self.RaiseError("In transition, type mismatch for state %s\n" % absolutePath.ToString())
      return S2MLFlattenerAction.FAILED
    sourceTargetStatePath = sourceTransition.GetTargetStatePath()
    targetTargetStatePath = S2MLPath.Copy(sourceTargetStatePath)
    targetTargetState = targetContainer.AccessNamedElement(targetTargetStatePath)
    if targetTargetState==None:
      self.RaiseError("In transition, undeclared state %s\n" % targetTargetStatePath.ToString())
      return S2MLFlattenerAction.FAILED
    if not AndreyModel.IsMarkovChainState(targetTargetState):
      absolutePath = targetTargetState.GetAbsolutePath()
      self.RaiseError("In transition, type mismatch for state %s\n" % absolutePath.ToString())
      return S2MLFlattenerAction.FAILED
    targetTransition = self._targetModel.NewMarkovChainTransition(targetEventPath, targetSourceStatePath, targetTargetStatePath, targetContainer)
    targetTransition.SetEvent(targetEvent)
    targetTransition.SetSourceState(targetSourceState)
    targetTransition.SetTargetState(targetTargetState)
    return S2MLFlattenerAction.COMPLETED

# 4) Flattening of models
# -----------------------

  def FlattenElement(self, action):
    sourceElement = action.GetDatum(1)
    if AndreyModel.IsMarkovChainState(sourceElement):
      status = self._FlattenMarkovChainState(action)
    elif AndreyModel.IsMarkovChainEvent(sourceElement):
      status = self._FlattenMarkovChainEvent(action)
    elif AndreyModel.IsMarkovChainParameter(sourceElement):
      status = self._FlattenMarkovChainParameter(action)
    elif AndreyModel.IsMarkovChainReward(sourceElement):
      status = self._FlattenMarkovChainParameter(action)
    elif AndreyModel.IsMarkovChainTransition(sourceElement):
      status = self._FlattenMarkovChainTransition(action)
    else:
      status = S2MLFlattener.FlattenElement(self, action)
    return status

  def _FlattenMarkovChainState(self, action):
    sourceContainer = action.GetDatum(0)
    sourceState = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceState.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetState = targetContainer.AccessNamedElement(targetPath)
    if targetState==None:
      targetState = self._targetModel.NewMarkovChainState(targetPath, targetContainer)
      targetState.Rename(targetPath.ToString())
    for name in sourceState.GetAttributeNames():
      sourceExpression = sourceState.GetAttribute(name)
      targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetState.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenMarkovChainEvent(self, action):
    sourceContainer = action.GetDatum(0)
    sourceEvent = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceEvent.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetEvent = targetContainer.AccessNamedElement(targetPath)
    if targetEvent==None:
      targetEvent = self._targetModel.NewMarkovChainEvent(targetPath, targetContainer)
      targetEvent.Rename(targetPath.ToString())
    for name in sourceEvent.GetAttributeNames():
      sourceExpression = sourceEvent.GetAttribute(name)
      targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetEvent.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenMarkovChainParameter(self, action):
    sourceContainer = action.GetDatum(0)
    sourceParameter = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceParameter.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetParameter = targetContainer.AccessNamedElement(targetPath)
    if targetParameter==None:
      targetParameter = self._targetModel.NewMarkovChainParameter(targetPath, targetContainer)
      targetParameter.SetSubType(sourceParameter.GetSubType())
      targetParameter.Rename(targetPath.ToString())
    sourceExpression = sourceParameter.GetDefinition()
    targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetParameter.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenMarkovChainTransition(self, action):
    sourceContainer = action.GetDatum(0)
    sourceTransition = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    # Event
    sourceEvent = sourceTransition.GetEvent()
    targetEventPath = sourceEvent.GetAbsolutePath()
    targetEventPath.RemoveFirstIdentifier()
    targetEventPath.MakeIdentifier()
    targetEvent = targetContainer.AccessNamedElement(targetEventPath)
    if targetEvent==None:
      self.RaiseError("No event %s: this should not happen" % targetEventPath.ToString())
      return S2MLFlattenerAction.FAILED
    # Source state
    sourceSourceState = sourceTransition.GetSourceState()
    targetSourceStatePath = sourceSourceState.GetAbsolutePath()
    targetSourceStatePath.RemoveFirstIdentifier()
    targetSourceStatePath.MakeIdentifier()
    targetSourceState = targetContainer.AccessNamedElement(targetSourceStatePath)
    if targetSourceState==None:
      self.RaiseError("No source state %s: this should not happen" % targetSourceStatePath.ToString())
      return S2MLFlattenerAction.FAILED
    # Target state
    sourceTargetState = sourceTransition.GetTargetState()
    targetTargetStatePath = sourceTargetState.GetAbsolutePath()
    targetTargetStatePath.RemoveFirstIdentifier()
    targetTargetStatePath.MakeIdentifier()
    targetTargetState = targetContainer.AccessNamedElement(targetTargetStatePath)
    if targetTargetState==None:
      self.RaiseError("No target state %s: this should not happen" % targetTargetStatePath.ToString())
      return S2MLFlattenerAction.FAILED
    # Transition
    targetTransition = self._targetModel.NewMarkovChainTransition(targetEventPath, targetSourceStatePath, targetTargetStatePath, targetContainer)
    targetTransition.SetEvent(targetEvent)
    targetTransition.SetSourceState(targetSourceState)
    targetTransition.SetTargetState(targetTargetState)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenStochasticExpression(self, sourceContainer, sourceExpression):
    if sourceExpression.GetType()==AndreyType.IN_EXPRESSION:
      sourceStateChild = sourceExpression.GetChild(0)
      sourceReferencePath = sourceStateChild.GetReferencePath()
      sourceReferredElement = sourceContainer.AccessNamedElement(sourceReferencePath)
      if sourceReferredElement==None or not AndreyModel.IsMarkovChainState(sourceReferredElement):
        absoluteSourcePath = sourceContainer.GetAbsolutePath()
        message = "In %s, wrong reference %s\n" % (absoluteSourcePath.ToString(), sourceReferencePath.ToString())
        self.RaiseError(message)
        return None
      targetReferencePath = sourceReferredElement.GetAbsolutePath()
      targetReferencePath.RemoveFirstIdentifier()
      targetStateChild = S2MLStochasticExpression.NewReference(targetReferencePath)
      sourceValueChild = sourceExpression.GetChild(1)
      targetValueChild = self._FlattenStochasticExpression(sourceContainer, sourceValueChild)
      if targetValueChild==None:
        return None
      return S2MLStochasticExpression.NewOperation(AndreyType.IN_EXPRESSION, [targetStateChild, targetValueChild])
    return S2MLFlattener._FlattenStochasticExpression(self, sourceContainer, sourceExpression)
