# Andrey version 1.0.0
# AndreyModels.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Andrey, a pedagogical implementation of Markov chains.
# Andrey is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Andrey is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Andrey.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) S2ML constructs
#  2.1) Blocks & classes
#  2.2) Instances
#  2.3) Extends directive
#  2.4) Clones directive
#  2.5) Embeds directive
# 3) Markov chains
#  3.1) Markov blocks & classes
#  3.2) States
#  3.3) Events
#  3.4) Parameters & rewards
#  3.5) Transitions
#  3.6) Models

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLModels import *
from AndreyType import *

# 2) Markov chains
# ================

# 2.1) Markov chain blocks
# ------------------------

class AndreyBlock(S2MLBlock):
  def __init__(self, identifier, container):
    S2MLBlock.__init__(self, identifier, container)
    self._subType = AndreyType.UNKNOWN
    self._timeType = AndreyType.UNKNOWN
    self._startIndexStates = -1
    self._startIndexImmediateStates = -1
    self._endIndexStates = -1
    self._startIndexTransitions = -1
    self._startIndexImmediateTransitions = -1
    self._endIndexTransitions = -1
    self._maximumRate = 0.0

  def GetSubType(self):
    return self._subType
  def SetSubType(self, subType):
    self._subType = subType

  def GetTimeType(self):
    return self._timeType
  def SetTimeType(self, type):
    self._timeType = type

  def GetStartIndexStates(self):
    return self._startIndexStates 
  def SetStartIndexStates(self, startIndexStates):
    self._startIndexStates = startIndexStates

  def GetStartIndexImmediateStates(self):
    return self._startIndexImmediateStates 
  def SetStartIndexImmediateStates(self, startIndexImmediateStates):
    self._startIndexImmediateStates = startIndexImmediateStates
  def HasImmediateState(self):
    return self.GetStartIndexImmediateStates()!=self.GetEndIndexStates()

  def GetEndIndexStates(self):
    return self._endIndexStates 
  def SetEndIndexStates(self, endIndexStates):
    self._endIndexStates = endIndexStates

  def GetStartIndexTransitions(self):
    return self._startIndexTransitions 
  def SetStartIndexTransitions(self, startIndexTransitions):
    self._startIndexTransitions = startIndexTransitions

  def GetStartIndexImmediateTransitions(self):
    return self._startIndexImmediateTransitions 
  def SetStartIndexImmediateTransitions(self, startIndexImmediateTransitions):
    self._startIndexImmediateTransitions = startIndexImmediateTransitions

  def GetEndIndexTransitions(self):
    return self._endIndexTransitions 
  def SetEndIndexTransitions(self, endIndexTransitions):
    self._endIndexTransitions = endIndexTransitions

  def GetMaximumRate(self):
    return self._maximumRate
  def SetMaximumRate(self, maximumRate):
    self._maximumRate = maximumRate

  def IsDiscrete(self):
    for element in self.GetElements():
      if AndreyModel.IsBlock(element):
        if not element.IsDiscrete():
          return False
      elif AndreyModel.IsMarkovChainTransition(element):
        if element.GetTimeType()==AndreyType.TEMPORIZED:
          return False
    return True

  def IsContinuous(self):
    isContinuous = False
    for element in self.GetElements():
      if AndreyModel.IsBlock(element):
        if not element.IsContinuous():
          return False
      elif AndreyModel.IsMarkovChainTransition(element):
        if element.GetTimeType()==AndreyType.TEMPORIZED:
          isContinuous = True
    return isContinuous

# 2.2) Markov chain classes
# -------------------------

class AndreyClass(S2MLClass):
  def __init__(self, identifier, container):
    S2MLClass.__init__(self, identifier, container)
    self._subType = AndreyType.UNKNOWN

  def GetSubType(self):
    return self._subType
  def SetSubType(self, subType):
    self._subType = subType

# 2.3) States
# -----------

class AndreyMarkovChainState(S2MLElement):
  def __init__(self, path, container):
    S2MLElement.__init__(self, AndreyType.MARKOV_CHAIN_STATE, container)
    container.SetSubType(AndreyType.MARKOV_CHAIN)
    self.SetPath(path)
    self._index = -1
    self._timeType = AndreyType.UNKNOWN
    self._probability = None

  def GetIndex(self):
    return self._index
  def SetIndex(self, index):
    self._index = index

  def GetTimeType(self):
    return self._timeType
  def SetTimeType(self, type):
    self._timeType = type

  def GetProbability(self):
    return self._probability
  def SetProbability(self, probability):
    self._probability = probability

# 2.4) Events
# -----------

class AndreyMarkovChainEvent(S2MLElement):
  def __init__(self, path, container):
    S2MLElement.__init__(self, AndreyType.MARKOV_CHAIN_EVENT, container)
    container.SetSubType(AndreyType.MARKOV_CHAIN)
    self.SetPath(path)
    self._timeType = AndreyType.UNKNOWN
    self._definition = None
    self._rate = -1
    self._probability = 0.0

  def GetTimeType(self):
    return self._timeType
  def SetTimeType(self, type):
    self._timeType = type

  def GetRate(self):
    return self._rate
  def SetRate(self, rate):
    self._rate = rate

  def GetProbability(self):
    return self._probability
  def SetProbability(self, probability):
    self._probability = probability

# 2.5) Parameters & rewards
# -------------------------

class AndreyMarkovChainParameter(S2MLParameter):
  def __init__(self, path, container):
    S2MLParameter.__init__(self, path, container)
    self._subType = AndreyType.UNKNOWN
    self._index = -1

  def GetSubType(self):
    return self._subType
  def SetSubType(self, subType):
    self._subType = subType

  def GetIndex(self):
    return self._index
  def SetIndex(self, index):
    self._index = index

# 2.6) Transitions
# ----------------

class AndreyMarkovChainTransition(S2MLElement):
  def __init__(self, eventPath, sourceStatePath, targetStatePath, container):
    S2MLElement.__init__(self, AndreyType.MARKOV_CHAIN_TRANSITION, container)
    container.SetSubType(AndreyType.MARKOV_CHAIN)
    self._eventPath = eventPath
    self._event = None
    self._sourceStatePath = sourceStatePath
    self._sourceState = None
    self._targetStatePath = targetStatePath
    self._targetState = None
    self._rate = 0.0
    self._probability = 0.0

  def GetEventPath(self):
    return self._eventPath
  def GetEvent(self):
    return self._event
  def SetEvent(self, event):
    self._event = event

  def GetTimeType(self):
    if self._event==None:
      return AndreyType.UNKNOWN
    return self._event.GetTimeType()

  def GetSourceStatePath(self):
    return self._sourceStatePath
  def SetSourceStatePath(self, sourceStatePath):
     self._sourceStatePath = sourceStatePath
  def GetSourceState(self):
    return self._sourceState
  def SetSourceState(self, sourceState):
    self._sourceState = sourceState

  def GetTargetStatePath(self):
    return self._targetStatePath
  def SetTargetStatePath(self, targetStatePath):
     self._targetStatePath = targetStatePath
  def GetTargetState(self):
    return self._targetState
  def SetTargetState(self, targetState):
    self._targetState = targetState

  def GetIndex(self):
    return self._index
  def SetIndex(self, index):
    self._index = index

  def GetRate(self):
    return self._rate
  def SetRate(self, rate):
    self._rate = rate

  def GetProbability(self):
    return self._probability
  def SetProbability(self, probability):
    self._probability = probability

# 2.7) Models
# -----------

class AndreyModel(S2MLModel):

  # Creator
  # -------

  def __init__(self):
    S2MLModel.__init__(self)

  # Methods to test types
  # ---------------------

  @staticmethod
  def IsMarkovChainBlock(element):
    return element.GetType()==AndreyType.BLOCK and element.GetSubType()==AndreyType.MARKOV_CHAIN
  @staticmethod
  def IsMarkovChainClass(element):
    return element.GetType()==AndreyType.CLASS and element.GetSubType()==AndreyType.MARKOV_CHAIN
  @staticmethod
  def IsMarkovSystemBlock(element):
    return element.GetType()==AndreyType.BLOCK and element.GetSubType()==AndreyType.MARKOV_SYSTEM
  @staticmethod
  def IsMarkovSystemClass(element):
    return element.GetType()==AndreyType.CLASS and element.GetSubType()==AndreyType.MARKOV_SYSTEM
  @staticmethod
  def IsMarkovChain(element):
    return AndreyModel.IsMarkovChainBlock(element) or AndreyModel.IsMarkovChainClass(element)
  @staticmethod
  def IsMarkovSystem(element):
    return AndreyModel.IsMarkovSystemBlock(element) or AndreyModel.IsMarkovSystemClass(element)

  @staticmethod
  def IsMarkovChainState(element):
    return element.GetType()==AndreyType.MARKOV_CHAIN_STATE
  @staticmethod
  def IsMarkovChainEvent(element):
    return element.GetType()==AndreyType.MARKOV_CHAIN_EVENT
  @staticmethod
  def IsMarkovChainParameter(element):
    return element.GetType()==AndreyType.PARAMETER and element.GetSubType()==AndreyType.MARKOV_CHAIN_PARAMETER
  @staticmethod
  def IsMarkovChainReward(element):
    return element.GetType()==AndreyType.PARAMETER and element.GetSubType()==AndreyType.MARKOV_CHAIN_REWARD
  @staticmethod
  def IsMarkovChainTransition(element):
    return element.GetType()==AndreyType.MARKOV_CHAIN_TRANSITION

  def IsValidReferredElementInStochasticExpression(self, element):
    return element.GetType()==AndreyType.PARAMETER or element.GetType()==AndreyType.MARKOV_CHAIN_STATE

  # Factory of elements
  # -------------------

  def NewBlock(self, identifier, container):
    return AndreyBlock(identifier, container)

  def NewClass(self, identifier, container):
    return AndreyClass(identifier, container)

  def NewMarkovChainState(self, path, container):
    return AndreyMarkovChainState(path, container)

  def NewMarkovChainEvent(self, path, container):
    return AndreyMarkovChainEvent(path, container)

  def NewMarkovChainParameter(self, path, container):
    return AndreyMarkovChainParameter(path, container)

  def NewMarkovChainTransition(self, eventPath, sourceStatePath, targetStatePath, container):
    return AndreyMarkovChainTransition(eventPath, sourceStatePath, targetStatePath, container)

  # Methods for the parser
  # ----------------------

  def CanComposeBlock(self, container):
    return S2MLModel.CanComposeBlock(self, container) and (container.GetType()==AndreyType.MODEL or container.GetSubType()==AndreyType.UNKNOWN or container.GetSubType()==AndreyType.MARKOV_SYSTEM)

  def CanComposeInstance(self, container):
    return S2MLModel.CanComposeInstance(self, container) and (container.GetSubType()==AndreyType.UNKNOWN or container.GetSubType()==AndreyType.MARKOV_SYSTEM)

  def CanComposeClonesDirective(self, container):
    return S2MLModel.CanComposeClonesDirective(self, container) and (container.GetSubType()==AndreyType.UNKNOWN or container.GetSubType()==AndreyType.MARKOV_SYSTEM)

  def CanComposeEmbedsDirective(self, container):
    return S2MLModel.CanComposeEmbedsDirective(self, container) and (container.GetSubType()==AndreyType.UNKNOWN or container.GetSubType()==AndreyType.MARKOV_SYSTEM)

  def CanComposeState(self, container):
    # return container.GetSubType()==AndreyType.UNKNOWN or container.GetSubType()==AndreyType.MARKOV_CHAIN or container.GetSubType()==AndreyType.MARKOV_SYSTEM
    return True

  def CanComposeEvent(self, container):
    # return container.GetSubType()==AndreyType.UNKNOWN or container.GetSubType()==AndreyType.MARKOV_CHAIN or container.GetSubType()==AndreyType.MARKOV_SYSTEM
    return True

  def CanComposeParameter(self, container):
    return True

  def CanComposeReward(self, container):
    return True

  def CanComposeTransition(self, container):
    return container.GetSubType()==AndreyType.UNKNOWN or container.GetSubType()==AndreyType.MARKOV_CHAIN

