# S2MLCore version 1.0.0
# S2MLFlattener.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore, a pedagogical calculation engine for Boolean reliability models.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Actions
# 3) Flattener
#   3.1) Creator
#   3.2) Error management
#   3.3) Compilation loop
#   3.4) Instantiation of models
#   3.5) Instantiation of directives
#   3.6) Instantiation of elements
#   3.7) Flattening of models

# 1) Imported modules
# -------------------

import string
import sys
import re
from S2MLMetaModel import *
from S2MLModels import *
from S2MLPrinter import *

# 2) Actions
# ----------

class S2MLFlattenerAction:
  INSTANTIATE_ELEMENT = 1
  FLATTEN_ELEMENT = 2

  TODO = 1
  POSTPONED = 2
  COMPLETED = 3
  FAILED = 4

  def __init__(self, type, size):
    self._type = type
    self._status = S2MLFlattenerAction.TODO
    self._size = size
    self._data = [None for _ in range(0, size)]

  def GetType(self):
    return self._type

  def GetStatus(self):
    return self._status
  def SetStatus(self, status):
    self._status = status

  def GetSize(self):
    return self._size
  def GetDatum(self, index):
    if index<0 or index>=self.GetSize():
      return None
    return self._data[index]
  def SetDatum(self, index, datum):
    if index<0 or index>=self.GetSize():
      return
    self._data[index] = datum

  @staticmethod
  def NewInstantiateElement(sourceElement, targetContainer):
    self = S2MLFlattenerAction(S2MLFlattenerAction.INSTANTIATE_ELEMENT, 2)
    self.SetDatum(0, sourceElement)
    self.SetDatum(1, targetContainer)
    return self

  @staticmethod
  def NewFlattenElement(sourceContainer, sourceElement, targetContainer):
    self = S2MLFlattenerAction(S2MLFlattenerAction.FLATTEN_ELEMENT, 3)
    self.SetDatum(0, sourceContainer)
    self.SetDatum(1, sourceElement)
    self.SetDatum(2, targetContainer)
    return self

# 3) Flattener
# ------------

# 3.1) Creator
# ------------

class S2MLFlattener:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._printer = self._engine.GetPrinter()
    self._sourceModel = None
    self._targetModel = None
    self._actionStack = []
    self._errors = 0
    self._noPostponing = False

# 3.2) Error management
# ---------------------

  def GetNumberOfErrors(self):
    return self._errors

  def RaiseError(self, message):
    fullMessage = ""
    self._errors += 1
    if self._errors>1:
      fullMessage += "\n"
    fullMessage += "Error %d\n" % self._errors
    fullMessage += message
    self._engine.RaiseError(message)

# 3.3) Compilation loop
# ---------------------

  def _IsActionStackEmpty(self):
    return len(self._actionStack)==0

  def _PushAction(self, action):
    self._actionStack.insert(0, action)

  def _EnqueueAction(self, action):
    self._actionStack.append(action)

  def _DequeueStack(self):
    return self._actionStack.pop(0)

  def _ExecuteCompilationLoop(self):
    firstPostponedAction = None
    self._noPostponing = False
    while not self._IsActionStackEmpty():
      action = self._DequeueStack()
      status = self._PerformAction(action)
      if status==S2MLFlattenerAction.COMPLETED:
        firstPostponedAction = None
      elif status==S2MLFlattenerAction.POSTPONED:
        self._EnqueueAction(action)
        if action==firstPostponedAction:
          break
        else:
          action.SetStatus(S2MLFlattenerAction.POSTPONED)
          if firstPostponedAction==None:
            firstPostponedAction = action
      elif status==S2MLFlattenerAction.FAILED:
        break
    if self._errors>0:
      return
    self._noPostponing = True
    while not self._IsActionStackEmpty():
      action = self._DequeueStack()
      status = self._PerformAction(action)
      if status==S2MLFlattenerAction.FAILED:
        break

  def _PerformAction(self, action):
    if action.GetType()==S2MLFlattenerAction.INSTANTIATE_ELEMENT:
      return self.InstantiateElement(action)
    elif action.GetType()==S2MLFlattenerAction.FLATTEN_ELEMENT:
      return self.FlattenElement(action)
    return S2MLFlattenerAction.FAILED

# 3.4) Instantiation of models
# ----------------------------

  def NewModel(self):
    return S2MLModel()

  def InstantiateModel(self, sourceModel):
    self._sourceModel = sourceModel
    self._targetModel = self.NewModel()
    self._errors = 0
    action = S2MLFlattenerAction.NewInstantiateElement(self._sourceModel, self._targetModel)
    self._PushAction(action)
    self._ExecuteCompilationLoop()
    if self._errors!=0:
      return None
    return self._targetModel

  def InstantiateAndFlattenModel(self, sourceModel):
    self._sourceModel = sourceModel
    self._targetModel = self.NewModel()
    self._errors = 0
    action = S2MLFlattenerAction.NewInstantiateElement(self._sourceModel, self._targetModel)
    self._PushAction(action)
    self._ExecuteCompilationLoop()
    if self._errors!=0:
      return None
    self._sourceModel = self._targetModel
    self._targetModel = self.NewModel()
    action = S2MLFlattenerAction.NewFlattenElement(self._sourceModel, self._sourceModel, self._targetModel)
    self._PushAction(action)
    self._ExecuteCompilationLoop()
    if self._errors!=0:
      return None
    return self._targetModel

  def InstantiateElement(self, action):
    sourceElement = action.GetDatum(0)
    if sourceElement.GetType()==S2MLType.MODEL:
      status = self._InstantiateModel(action)
    elif sourceElement.GetType()==S2MLType.DOMAIN:
      status = self._InstantiateDomain(action)
    elif sourceElement.GetType()==S2MLType.BLOCK:
      status = self._InstantiateBlock(action)
    elif sourceElement.GetType()==S2MLType.CLASS:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.INSTANCE:
      status = self._InstantiateInstance(action)
    elif sourceElement.GetType()==S2MLType.CLONES:
      status = self._InstantiateClonesDirective(action)
    elif sourceElement.GetType()==S2MLType.EXTENDS:
      status = self._InstantiateExtendsDirective(action)
    elif sourceElement.GetType()==S2MLType.EMBEDS:
      status = self._InstantiateEmbedsDirective(action)
    elif sourceElement.GetType()==S2MLType.PARAMETER:
      status = self._InstantiateParameter(action)
    else:
      status = S2MLFlattenerAction.FAILED
    return status

  def _PushInstantiateActionsForElements(self, sourceContainer, targetContainer):
    for sourceElement in reversed(sourceContainer.GetElements()):
      self._PushInstantiateActionForElement(sourceElement, targetContainer)

  def _PushInstantiateActionForElement(self, sourceElement, targetContainer):
    action = S2MLFlattenerAction.NewInstantiateElement(sourceElement, targetContainer)
    self._PushAction(action)

  def _InstantiateModel(self, action):
    sourceModel = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    self._PushInstantiateActionsForElements(sourceModel, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateDomain(self, action):
    sourceDomain = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetDomain = self._targetModel.NewDomain(sourceDomain.GetIdentifier(), targetContainer)
    for sourceConstant in sourceDomain.GetConstants():
      targetConstant = self._targetModel.NewSymbol(sourceConstant.GetValue())
      targetDomain.AppendConstant(targetConstant)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateBlock(self, action):
    sourceBlock = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetBlock = self._targetModel.NewBlock(sourceBlock.GetIdentifier(), targetContainer)
    self._PushInstantiateActionsForElements(sourceBlock, targetBlock)
    return S2MLFlattenerAction.COMPLETED

# 3.5) Instantiation of directives
# --------------------------------

  def _InstantiateInstance(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceClassIdentifier = sourceDirective.GetClassIdentifier()
    sourceClass = self._sourceModel.LookForClass(sourceClassIdentifier)
    if sourceClass==None:
      parentBlock = sourceDirective.GetContainer()
      absoluteSourcePath = parentBlock.GetAbsolutePath()
      message = "In %s, attempt to instantiate undefined class %s\n" % (absoluteSourcePath.ToString(), sourceClassIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetIdentifier = sourceDirective.GetIdentifier()
    targetBlock = targetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetBlock!=None:
      parentBlock = sourceDirective.GetContainer()
      absoluteSourcePath = parentBlock.GetAbsolutePath()
      message = "In %s, element %s already exists\n" % (absoluteSourcePath.ToString(), targetIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetBlock = self._targetModel.NewBlock(targetIdentifier, targetContainer)
    self._PushInstantiateActionsForElements(sourceDirective, targetBlock)
    self._PushInstantiateActionsForElements(sourceClass, targetBlock)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateExtendsDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceClassIdentifier = sourceDirective.GetClassIdentifier()
    sourceClass = self._sourceModel.LookForClass(sourceClassIdentifier)
    if sourceClass==None:
      message = "In %s, attempt to inherit undefined class %s\n" % (sourceContainerAbsolutePath.ToString(), sourceClassIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    self._PushInstantiateActionsForElements(sourceClass, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateClonesDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    pathToClonedElement = sourceDirective.GetPathToClonedElement()
    sourceBlock = targetContainer.AccessNamedElement(pathToClonedElement)
    if sourceBlock==None:
      message = "In %s, attempt to clone undefined block %s\n" % (absoluteSourcePath.ToString(), pathToClonedElement.ToString())
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetIdentifier = sourceDirective.GetIdentifier()
    targetBlock = targetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetBlock!=None:
      absoluteSourcePath = sourceDirective.GetAbsolutePath()
      message = "In %s, element %s already exists\n" % (absoluteSourcePath.ToString(), targetIdentifier)
      self.RaiseError(message)
      return S2MLFlattenerAction.FAILED
    targetBlock = self._targetModel.NewBlock(targetIdentifier, targetContainer)
    self._PushInstantiateActionsForElements(sourceDirective, targetBlock)
    self._PushInstantiateActionsForElements(sourceBlock, targetBlock)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateEmbedsDirective(self, action):
    sourceDirective = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetIdentifier = sourceDirective.GetIdentifier()
    targetBlock = targetContainer.LookForBlock(targetIdentifier)
    if targetBlock!=None:
      absoluteSourcePath = sourceDirective.GetAbsolutePath()
      message = "In %s, element %s already exists\n" % (absoluteSourcePath.ToString(), targetIdentifier)
      self.RaiseError(message)
    pathToEmbeddedElement = sourceDirective.GetPathToEmbeddedElement()
    targetContainer.SetAlias(targetIdentifier, pathToEmbeddedElement)
    return S2MLFlattenerAction.COMPLETED

# 3.6) Instantiation of elements
# ------------------------------

  def _InstantiateParameter(self, action):
    sourceParameter = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceParameter.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        targetParameter = self._targetModel.NewParameter(targetPath, targetContainer)
        targetParameter.Rename(targetPath.ToString())
        contextualizedPath = S2MLPath.NewEmptyPath()
      else:
        absoluteSourcePath = sourceParameter.GetAbsolutePath()
        self.RaiseError("Parameter %s declared outside of its parent block\n" % absoluteSourcePath.ToString())
        return S2MLFlattenerAction.FAILED
    elif targetElement.GetType()==S2MLType.PARAMETER:
      targetParameter = targetElement # self._targetModel.NewParameter(targetPath, targetContainer)
      contextualizedPath = targetContainer.GetContextualizedAbsolutePath()
    else:
      absoluteSourcePath = sourceParameter.GetAbsolutePath()
      self.RaiseError("Type mismatch for parameter %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    sourceExpression = sourceParameter.GetDefinition()
    targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, contextualizedPath)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetParameter.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _ContextualizedCopyOfStochasticExpression(self, expression, contextualizedPath):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      sourceReferencePath = expression.GetReferencePath()
      targetReferencePath = S2MLPath.Merge(contextualizedPath, sourceReferencePath)
      return S2MLStochasticExpression.NewReference(targetReferencePath)
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      value = expression.GetValue()
      return S2MLStochasticExpression.NewConstant(value)
    childrenCopy = self._ContextualizedCopyOfChildrenOfStochasticExpression(expression, contextualizedPath)
    if childrenCopy==None:
      return None
    return S2MLStochasticExpression.NewOperation(expression.GetType(), childrenCopy)

  def _ContextualizedCopyOfChildrenOfStochasticExpression(self, expression, contextualizedPath):
    childrenCopy = []
    for child in expression.GetChildren():
      childCopy = self._ContextualizedCopyOfStochasticExpression(child, contextualizedPath)
      if childCopy==None:
        return None
      childrenCopy.append(childCopy)
    return childrenCopy

# 3.7) Flattening of models
# -------------------------

  def FlattenElement(self, action):
    sourceElement = action.GetDatum(1)
    if sourceElement.GetType()==S2MLType.MODEL:
      status = self._FlattenModel(action)
    elif sourceElement.GetType()==S2MLType.DOMAIN:
      status = self._FlattenDomain(action)
    elif sourceElement.GetType()==S2MLType.BLOCK:
      status = self._FlattenBlock(action)
    elif sourceElement.GetType()==S2MLType.CLASS:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.INSTANCE:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.CLONES:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.EXTENDS:
      status = S2MLFlattenerAction.COMPLETED
    elif sourceElement.GetType()==S2MLType.PARAMETER:
      status = self._FlattenParameter(action)
    else:
      status = S2MLFlattenerAction.FAILED
    return status

  def _FlattenModel(self, action):
    sourceModel = action.GetDatum(0)
    targetContainer = action.GetDatum(2)
    self._PushFlattenActionsForElements(sourceModel, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _PushFlattenActionsForElements(self, sourceContainer, targetContainer):
    for sourceElement in reversed(sourceContainer.GetElements()):
      self._PushFlattenActionForElement(sourceContainer, sourceElement, targetContainer)

  def _PushFlattenActionForElement(self, sourceContainer, sourceElement, targetContainer):
    action = S2MLFlattenerAction.NewFlattenElement(sourceContainer, sourceElement,  targetContainer)
    self._PushAction(action)

  def _FlattenDomain(self, action):
    sourceContainer = action.GetDatum(0)
    sourceDomain = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetDomain = self._targetModel.NewDomain(sourceDomain.GetIdentifier(), targetContainer)
    for sourceConstant in sourceDomain.GetConstants():
      targetConstant = self._targetModel.NewSymbol(sourceConstant.GetValue())
      targetDomain.AppendConstant(targetConstant)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenBlock(self, action):
    sourceContainer = action.GetDatum(0)
    sourceBlock = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    if sourceContainer==self._sourceModel:
      targetBlock = self._targetModel.NewBlock(sourceBlock.GetIdentifier(), targetContainer)
      self._PushFlattenActionsForElements(sourceBlock, targetBlock)
    else:
      self._PushFlattenActionsForElements(sourceBlock, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenParameter(self, action):
    sourceContainer = action.GetDatum(0)
    sourceParameter = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourcePath = sourceParameter.GetPath()
    actualSourceParameter = sourceContainer.AccessNamedElement(sourcePath)
    if actualSourceParameter==None:
      self.RaiseError("This should not happen")
      return assisFlattenerAction.FAILED
    targetPath = actualSourceParameter.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetIdentifier = targetPath.ToString()
    targetParameter = targetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetParameter==None:
      targetParameter = self._targetModel.NewParameter(targetPath, targetContainer)
      targetParameter.Rename(targetIdentifier)
    sourceExpression = sourceParameter.GetDefinition()
    targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetParameter.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenStochasticExpression(self, sourceContainer, sourceExpression):
    if sourceExpression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      sourceReferencePath = sourceExpression.GetReferencePath()
      sourceReferredElement = sourceContainer.AccessNamedElement(sourceReferencePath)
      #if sourceReferredElement==None or not S2MLModel.IsParameter(sourceReferredElement):
      if sourceReferredElement==None:
        absoluteSourcePath = sourceContainer.GetAbsolutePath()
        message = "In %s, undefined reference %s\n" % (absoluteSourcePath.ToString(), sourceReferencePath.ToString())
        self.RaiseError(message)
        return None
      targetReferencePath = sourceReferredElement.GetAbsolutePath()
      targetReferencePath.RemoveFirstIdentifier()
      return S2MLStochasticExpression.NewReference(targetReferencePath)
    if sourceExpression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      value = sourceExpression.GetValue()
      return S2MLStochasticExpression.NewConstant(value)
    targetChildren = self._FlattenChildrenOfStochasticExpression(sourceContainer, sourceExpression)
    if targetChildren==None:
      return None
    return S2MLStochasticExpression.NewOperation(sourceExpression.GetType(), targetChildren)

  def _FlattenChildrenOfStochasticExpression(self, sourceContainer, sourceExpression):
    targetChildren = []
    for sourceChild in sourceExpression.GetChildren():
      targetChild = self._FlattenStochasticExpression(sourceContainer, sourceChild)
      if targetChild==None:
        return None
      targetChildren.append(targetChild)
    return targetChildren
