# S2MLCore version 1.0.0
# S2MLMetaModel.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Paths
# 3) Elements
# 4) Containers
# 5) Terms

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

# 2) Paths
# --------

class S2MLPath:
  COMPOSITION_SEPARATOR = "."
  PSEUDO_COMPOSITION_SEPARATOR = "_"
  ROOT_IDENTIFIER = "main"
  CONTAINER_IDENTIFIER = "owner"

  def __init__(self):
    self._identifiers = []
    self._string = None

  @staticmethod
  def NewEmptyPath():
    return S2MLPath()
    
  @staticmethod
  def FromString(string):
    if string=="":
      return S2MLPath.NewEmptyPath()
    identifiers = string.split(S2MLPath.COMPOSITION_SEPARATOR)
    for identifier in identifiers:
      if not S2MLPath.IsValidIdentifier(identifier):
        return None
    self = S2MLPath()
    self._identifiers = identifiers
    self._string = string
    return self

  @staticmethod
  def Copy(path):
    self = S2MLPath()
    for identifier in path._identifiers:
      self.AppendIdentifier(identifier)
    if path._string!=None:
      self._string = path._string
    return self

  @staticmethod
  def Merge(path1, path2):
    self = S2MLPath()
    self._identifiers = path1._identifiers + path2._identifiers
    return self

  @staticmethod
  def IsValidIdentifier(string):
    return re.match("^[a-zA-Z_][a-zA-Z0-9_-]*$", string)

  @staticmethod
  def IsValidPath(string):
    identifiers = string.split(S2MLPath.COMPOSITION_SEPARATOR)
    for identifier in identifiers:
      if not S2MLPath.IsValidIdentifier(identifier):
        return False
    return True

  def ToString(self):
    if self._string!=None:
      return self._string
    self._string = ""
    for identifier in self._identifiers:
      if self._string=="":
        self._string = identifier
      else:
        self._string += S2MLPath.COMPOSITION_SEPARATOR + identifier
    return self._string

  def ToIdentifier(self):
    string = ""
    for identifier in self._identifiers:
      if string=="":
        string = identifier
      else:
        string += S2MLPath.PSEUDO_COMPOSITION_SEPARATOR + identifier
    return string

  def MakeIdentifier(self):
    identifier = self.ToString()
    self._string = identifier
    self._identifiers = [identifier]

  def IsEmpty(self):
    return self.GetSize()==0
  def GetSize(self):
    return len(self._identifiers)

  def IsIdentifier(self):
    return self.GetSize()==1

  def GetLastIdentifier(self):
    return self.GetIdentifier(self.GetSize()-1)

  def GetFirstIdentifier(self):
    return self.GetIdentifier(0)

  def GetIdentifier(self, index):
    if index<0 or index>=len(self._identifiers):
      return None
    return self._identifiers[index]

  def AppendIdentifier(self, identifier):
    self._identifiers.append(identifier)
    self._string = None

  def PrependIdentifier(self, identifier):
    self._identifiers.insert(0, identifier)
    self._string = None

  def RemoveLastIdentifier(self):
    self._string = None
    return self._identifiers.pop()

  def RemoveFirstIdentifier(self):
    self._string = None
    return self._identifiers.pop(0)

  def RemoveLastIdentifier(self):
    self._string = None
    return self._identifiers.pop()

  def GetIdentifiers(self):
    return self._identifiers

  def Resolve(self):
    oldIdentifiers = self._identifiers
    self._identifiers = []
    for identifier in oldIdentifiers:
      if identifier==S2MLPath.ROOT_IDENTIFIER:
        self._identifiers = []
        self._identifiers.append(S2MLPath.ROOT_IDENTIFIER)
      elif identifier==S2MLPath.CONTAINER_IDENTIFIER:
        if len(self._identifiers)==0:
          self._identifiers = oldIdentifiers
          return False
          self._identifiers.pop()
      else:
        self._identifiers.append(identifier)
    return True

  def GetSuffix(self, other):
    if len(self._identifiers)<len(other._identifiers):
      return None
    for i in range(0, len(other._identifiers)):
      if self.GetIdentifier(i)!=other.GetIdentifier(i):
        return None
    result = S2MLPath()
    for i in range(len(other._identifiers), len(self._identifiers)):
      result.AppendIdentifier(self.GetIdentifier(i))
    return result

# 3) Elements
# -----------

class S2MLElement:
  def __init__(self, type, container):
    self._type = type
    self._isAtomic = True
    self._identifier = None
    self._path = None
    self._container = container
    if container!=None:
      container._AddElement(self)
    self._definition = None
    self._flag = 0
    self._attributes = dict()

  def Delete(self):
    if self._container==None:
      return
    if self._identifier!=None:
      self._container._RemoveFromDictonary(self)
    self._container._RemoveElement(self)

  def GetType(self):
    return self._type

  def IsAtomic(self):
    return self._isAtomic

  def GetIdentifier(self):
    return self._identifier
  def SetIdentifier(self, identifier):
    self._identifier = identifier
  def Rename(self, identifier):
    if self._container==None:
      self._identifier = identifier
      return
    if self._identifier!=None:
      self._container._RemoveFromDictonary(self)
    self._identifier = identifier
    self._container._AddToDictonary(self)

  def GetPath(self):
    return self._path
  def SetPath(self, path):
    self._path = path

  def GetAbsolutePath(self):
    path = S2MLPath.Copy(self.GetPath())
    container = self.GetContainer()
    if container!=None:
      while container.GetContainer()!=None:
        path.PrependIdentifier(container.GetIdentifier())
        container = container.GetContainer()
    return path

  def GetContextualizedAbsolutePath(self):
    path = self.GetAbsolutePath()
    path.RemoveFirstIdentifier()
    path.PrependIdentifier(S2MLPath.ROOT_IDENTIFIER)
    return path

  def GetContainer(self):
    return self._container
  def GetModel(self):
    if self.GetContainer()==None:
      return self
    return self.GetContainer().GetModel()
  def GetRootContainer(self):
    container = self.GetContainer()
    if container==None:
      return None
    parentContainer = container.GetContainer()
    if parentContainer==None:
      return self
    return container.GetRootContainer()

  def GetDefinition(self):
    return self._definition
  def SetDefinition(self, definition):
    self._definition = definition

  def GetFlag(self):
    return self._flag
  def SetFlag(self, value):
    self._flag = value

  def GetNumberOfAttributes(self):
    return len(self._attributes)
  def GetAttribute(self, name, defaultValue = None):
    return self._attributes.get(name, defaultValue)
  def SetAttribute(self, name, value):
    self._attributes[name] = value
  def ResetAttribute(self, name):
    self._attributes.pop(name, None)
  def ResetAttributes(self):
    self._attributes.clear()
  def GetAttributeNames(self):
    return self._attributes

# 4) Containers
# -------------

class S2MLContainer(S2MLElement):
  def __init__(self, type, container):
    S2MLElement.__init__(self, type, container)
    self._isAtomic = False
    self._elements = []
    self._dictionary = dict()
    self._aliases = dict()

  def DeleteAll(self):
    while self.GetNumberOfElements()>0:
      element = self.GetElement(0)
      element.Delete()
    self.Delete()

  def GetNumberOfElements(self):
    return len(self._elements)
  def GetElement(self, index):
    if index<0 or index>=self.GetNumberOfElements():
      return None
    return self._elements[index]
  def GetElements(self):
    return self._elements
  def GetDictonary(self):
    return self._dictionary
  def LookForIdentifiedElement(self, identifier, checkType = None):
    element = self._dictionary.get(identifier, None)
    if element==None:
      return None
    if checkType!=None and not checkType(element):
      return None
    return element
  def _AddElement(self, element):
    self._elements.append(element)
  def _RemoveElement(self, element):
    self._elements.remove(element)
  def _AddToDictonary(self, element):
    # REQUIRE: self.LookForIdentifiedElement(element.GetIdentifier())==None
    self._dictionary[element.GetIdentifier()] = element
  def _RemoveFromDictonary(self, element):
    del self._dictionary[element.GetIdentifier()]
  def AccessNamedElement(self, path):
    if path.IsEmpty():
      return self
    identifier = path.RemoveFirstIdentifier()
    if identifier==S2MLPath.ROOT_IDENTIFIER:
      container = self.GetRootContainer()
      element = container.AccessNamedElement(path)
    elif identifier==S2MLPath.CONTAINER_IDENTIFIER:
      container = self.GetContainer()
      if container==None:
        element = None
      else:
        element = container.AccessNamedElement(path)
    else:
      aliasPath = self.GetAlias(identifier)
      if aliasPath==None:
        element = self.LookForIdentifiedElement(identifier)
        if element!=None:
          if not element.IsAtomic():
            container = element
            element = container.AccessNamedElement(path)
          elif not path.IsEmpty():
            element = None
      else:
        aliasedPath = S2MLPath.Merge(aliasPath, path)
        element = self.AccessNamedElement(aliasedPath)
        if element==None:
          print("Nope")
    path.PrependIdentifier(identifier)
    return element
  def RemoveElement(self, element):
    if not element in self._elements:
      return -1
    index = self._elements.index(element)
    self._elements.remove(element)
    if element.GetIdentifier()!=None:
      self._RemoveFromDictonary(element)
    element._container = None
    return index
  def AddElement(self, element):
    self._elements.append(element)
    if element.GetIdentifier()!=None:
      self._AddToDictonary(element)
    element._container = self
  def InsertElement(self, index, element):
    self._elements.insert(index, element)
    if element.GetIdentifier()!=None:
      self._AddToDictonary(element)
    element._container = self

  def CollectElements(self, test, recursive, elements):
    for element in self.GetElements():
      if test(element):
        elements.append(element)
      if not element._isAtomic and recursive:
        element.CollectElements(test, recursive, elements)
    if test(self):
      elements.append(self)

  def GetAlias(self, identifier):
    return self._aliases.get(identifier, None)
  def SetAlias(self, identifier, path):
    self._aliases[identifier] = path

  def SetFlagsOfElements(self, value):
    for element in self._elements:
      element.SetFlag(value)
      if not element.IsAtomic():
        element.SetFlagsOfElements(value)
  def ResetFlagsOfElements(self):
    self.SetFlagsOfElements(0)

  def SetAttributesOfElements(self, testElement, attributeName, attributeValue, recursive=True):
    for element in self.GetElements():
      if testElement(element):
        element.SetAttribute(attributeName, attributeValue)
      if not element.IsAtomic() and recursive:
        element.SetAttributesOfElements(testElement, attributeName, attributeValue, True)
  def ResetAttributesOfElements(self, testElement, attributeName, recursive=True):
    for element in self.GetElements():
      if testElement(element):
        element.ResetAttribute(attributeName)
      if not element.IsAtomic() and recursive:
        element.ResetAttributesOfElements(testElement, attributeName, True)

# 5) Terms
# --------

class S2MLTerm:
  def __init__(self, type):
    self._type = type
    self._referencePath = None
    self._referredElement = None
    self._value = None
    self._children = []
    self._attributes = dict()

  def GetType(self):
    return self._type

  def GetReferencePath(self):
    return self._referencePath
  def SetReferencePath(self, path):
    self._referencePath = path

  def GetReferredElement(self):
    return self._referredElement
  def SetReferredElement(self, element):
    self._referredElement = element

  def GetValue(self):
    return self._value
  def SetValue(self, value):
    self._value = value

  def GetNumberOfChildren(self):
    return len(self._children)
  def GetChild(self, i):
    if i<0 or i>=len(self._children):
      return None
    return self._children[i]
  def GetChildren(self):
    return self._children
  def AppendChild(self, child):
    self._children.append(child)
  def AppendChildren(self, children):
    self._children.extend(children)

  def GetAttribute(self, name, defaultValue = None):
    return self._attributes.get(name, defaultValue)
  def SetAttribute(self, name, value):
    self._attributes[name] = value
  def ResetAttribute(self, name):
    self._attributes.pop(name, None)
  def ResetAttributes(self):
    self._attributes.clear()
