# S2MLCore version 1.0.0
# S2MLPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
# 4) Domains, blocks and classes
# 5) Directives
# 6) Elements
# 7) Parameters and stochastic expressions
# 8) Matrices

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLMatrices import *
from S2MLStochasticExpressions import *
from S2MLLexer import *

# 2) Creator
# ----------

class S2MLPrinter:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._lexer = engine.GetLexer()
    self._depth = 0
    self._indentation = "  "
    self._columnSeparator = "\t"
    self._floatFormat = "%g"

  def RaiseError(self, message):
    self._engine.RaiseError(message)
    
  def GetIndentation(self):
    return self._indentation
  def SetIndentation(self, indentation):
    self._indentation = indentation

  def GetColumnSeparator(self):
    return self._columnSeparator
  def SetColumnSeparator(self, columnSeparator):
    self._columnSeparator = columnSeparator

  def GetFloatFormat(self):
    return self._floatFormat
  def SetFloatFormat(self, floatFormat):
    self._floatFormat = floatFormat

  def PrintIndentation(self, output):
    for i in range(self._depth):
      output.write(self._indentation)

  def PrintSymbol(self, symbolCode, output):
    output.write(self._lexer.GetStringOfCode(symbolCode))

  def PrintKeyword(self, keywordCode, output):
    output.write(self._lexer.GetStringOfCode(keywordCode))

# 3) Models
# ---------

  def ExportModel(self, model, fileName, mode):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintModel(model, output)
    output.close()
    return 0

  def PrintModel(self, model, output):
    self._depth = 0
    for element in model.GetElements():
      self.PrintElement(element, output)

  # Must be derived
  def PrintElement(self, element, output):
    if element.GetType()==S2MLType.DOMAIN:
      self.PrintDomain(element, output)
    elif element.GetType()==S2MLType.BLOCK:
      self.PrintBlock(element, output)
    elif element.GetType()==S2MLType.CLASS:
      self.PrintClass(element, output)
    elif element.GetType()==S2MLType.INSTANCE:
      self.PrintInstance(element, output)
    elif element.GetType()==S2MLType.CLONES:
      self.PrintClonesDirective(element, output)
    elif element.GetType()==S2MLType.EXTENDS:
      self.PrintExtendsDirective(element, output)
    elif element.GetType()==S2MLType.EMBEDS:
      self.PrintEmbedsDirective(element, output)
    elif element.GetType()==S2MLType.PARAMETER:
      self.PrintParameter(element, output)

# 4) Domains, blocks and classes
# ------------------------------

  def PrintDomain(self, domain, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_DOMAIN, output)
    output.write(" ")
    output.write(domain.GetIdentifier())
    output.write(" {")
    first = True
    for constant in domain.GetConstants():
      if first:
        first = False
      else:
        output.write(", ")
      self.PrintConstant(constant, output)
    output.write("}\n")

  def PrintBlock(self, block, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_BLOCK, output)
    output.write(" ")
    output.write(block.GetIdentifier())
    output.write("\n")
    self._PrintBlockBody(block, output)
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_END, output)
    output.write("\n")

  def PrintClass(self, aClass, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_CLASS, output)
    output.write(" ")
    output.write(aClass.GetIdentifier())
    output.write("\n")
    self._PrintBlockBody(aClass, output)
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_END, output)
    output.write("\n")

  def _PrintBlockBody(self, block, output):
    self._depth += 1
    for element in block.GetElements():
      self.PrintElement(element, output)
    self._depth -= 1

# 5) Directives
# -------------

  def PrintInstance(self, directive, output):
    self.PrintIndentation(output)
    output.write(directive.GetClassIdentifier())
    output.write(" ")
    output.write(directive.GetIdentifier())
    if directive.GetNumberOfElements()==0:
      output.write(";\n")
    else:
      output.write("\n")
      self._depth += 1
      for element in directive.GetElements():
        self.PrintElement(element, output)
      self._depth -= 1
      self.PrintIndentation(output)
      self.PrintKeyword(S2MLType.KEYWORD_END, output)
      output.write("\n")

  def PrintClonesDirective(self, directive, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_CLONES, output)
    output.write(" ")
    self.PrintPath(directive.GetPathToClonedElement(), output)
    output.write(" as ")
    output.write(directive.GetIdentifier())
    if directive.GetNumberOfElements()==0:
      output.write(";\n")
    else:
      output.write("\n")
      self._depth += 1
      for element in directive.GetElements():
        self.PrintElement(element, output)
      self._depth -= 1
      self.PrintIndentation(output)
      self.PrintKeyword(S2MLType.KEYWORD_END, output)
      output.write("\n")

  def PrintExtendsDirective(self, directive, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_EXTENDS, output)
    output.write(" ")
    output.write(directive.GetClassIdentifier())
    output.write(";\n")

  def PrintEmbedsDirective(self, directive, output):
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_EMBEDS, output)
    output.write(" ")
    self.PrintPath(directive.GetPathToEmbeddedElement(), output)
    output.write(" ")
    self.PrintKeyword(S2MLType.KEYWORD_AS, output)
    output.write(" ")
    output.write(directive.GetIdentifier())
    if directive.GetNumberOfElements()==0:
      output.write(";\n")
    else:
      output.write("\n")
      self._depth += 1
      for element in directive.GetElements():
        self.PrintElement(element, output)
      self._depth -= 1
      self.PrintIndentation(output)
      self.PrintKeyword(S2MLType.KEYWORD_END, output)
      output.write("\n")

  def PrintPath(self, path, output):
    output.write(path.ToString())

# 6) Elements
# -----------

  def PrintAttributes(self, element, output):
    if element.GetNumberOfAttributes()==0:
      return
    output.write("(")
    first = True
    for name in element.GetAttributeNames():
      if first:
        first = False
      else:
        output.write(", ")
      output.write(name)
      output.write("=")
      value = element.GetAttribute(name)
      self.PrintStochasticExpression(value, output)
    output.write(")")

# 7) Parameters and stochastic expressions
# ----------------------------------------

  def PrintParameter(self, parameter, output):
    self._UpdateLocation(False, output)
    self.PrintIndentation(output)
    self.PrintKeyword(S2MLType.KEYWORD_PARAMETER, output)
    output.write(" ")
    self.PrintPath(parameter.GetPath(), output)
    output.write(" = ")
    self.PrintStochasticExpression(parameter.GetDefinition(), output)
    output.write(";\n")

  def PrintStochasticExpression(self, expression, output):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      self.PrintPath(expression.GetReferencePath(), output)
    elif expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      self.PrintConstant(expression.GetValue(), output)
    else:
      output.write("(")
      operator = self._engine.GetExpressionManager().GetOperatorOfType(expression.GetType())
      keywordCode = operator.GetKeywordCode()
      self.PrintKeyword(keywordCode, output)
      self.PrintChildrenOfStochasticExpression(expression, output)
      output.write(")")

  def PrintChildrenOfStochasticExpression(self, expression, output):
    for child in expression.GetChildren():
      output.write(" ")
      self.PrintStochasticExpression(child, output)

# 8) Matrices
# -----------

  def ExportMatrix(self, matrix, fileName, mode="w"):
    try:
      output = open(fileName, mode)
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return 1
    self.PrintMatrix(matrix, output)
    output.close()
    return 0

  def PrintMatrix(self, matrix, output):
    for rowIndex in range(matrix.GetNumberOfRows()):
      rowVector = matrix.GetRow(rowIndex)
      self.PrintVector(rowVector, output)
      output.write("\n")
    output.write("\n")
    output.flush()

  def PrintVector(self, vector, output):
    for index in range(vector.GetSize()):
     if index>0:
       output.write(self._columnSeparator)
     constant = vector.Get(index)
     self.PrintConstant(constant, output)

  def PrintConstant(self, constant, output):
    if constant.GetType()==S2MLConstant.NONE:
      pass
    elif constant.GetType()==S2MLConstant.BOOLEAN:
      if constant.GetValue():
        output.write("true")
      else:
        output.write("false")
    elif constant.GetType()==S2MLConstant.INTEGER:
      output.write("%d" % constant.GetValue())
    elif constant.GetType()==S2MLConstant.REAL:
      output.write(constant.ToString())
    else:
      output.write(constant.GetValue())
